/* iip.c - Inline Images Protocol */
/* Created: Sat Nov 18 09:41:36 AM EST 2023 malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* $Id: iip.c,v 1.1 2023/11/18 19:24:19 malakai Exp $ */

/* Copyright © 2023 Rahjiii <rahjiii@jeffrika.com>
 *
 * This file is part of iip - Last Outpost Inline Images Protocol
 *
 * iip is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * iip is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Locolor.  If not, see <https://www.gnu.org/licenses/>.
 */


/* This file implements the Inline Images Protocol for displaying images inside
 * of a terminal window.  See the documentation at
 * https://iterm2.com/documentation-images.html and
 * https://github.com/xtermjs/xterm.js/tree/master/addons/addon-image for info
 * on how the protocol works. */

/* ---- System includes ---- */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <glib.h>

/* ---- Integration includes ---- */
/* LO - structs.h is where my raw_output() descriptor type lives.  comm.c is where
 * my game's raw_output() function lives.  utils.h is where my game's error
 * logging function lives. */
#include "structs.h"
#include "comm.h"
#include "utils.h"

/* ---- Local includes ---- */
#include "iip.h"

/* ---- local #defines ---- */
/* the maximum allowable size of an image file, in bytes.*/
#define IIP_SIZE_LIMIT 20000000

/* ---- local structs and typedefs ---- */

/* ---- local variable declarations ---- */

/* This is an array that maps the width/height unit type into the appropriate
 * suffix defined by the IIP protocol. */
gchar *iip_unit_suffix[IIP_UNIT_MAX] = {
	[IIP_UNIT_NONE] = "",
	[IIP_UNIT_CELLS] = "",
	[IIP_UNIT_PIXELS] = "px",
	[IIP_UNIT_PERCENT] = "%",
	[IIP_UNIT_AUTO] = "auto"
};

/* ---- local function forward declarations ---- */
void iip_raw_output(gchar *msg, iip_descriptor_t d);
void iip_error(gchar *msg);

/* ---- CODE STARTS HERE ---- */

/* iip_raw_output() - sends a null terminated string to the provided
 * descriptor.  You are expected to change this function to match your
 * application's output requirements.  Note that for the protocol to work, the
 * output function you call must NOT send line termination (no CR or LF), and
 * it must be 8-bit clean, except for the '\0' character. It doesn't have to be
 * able to send a zero byte, but it does need to send other terminal escape
 * sequences.  See also the comment near the iip_descriptor_t typedef in iip.h.
 * */
void iip_raw_output(gchar *msg, iip_descriptor_t d) {
	/* My MUD game uses "raw_output(char *txt, struct descriptor_data *d)", but
	 * because gchar*'s are compatible with char*'s, and I've typedef'd
	 * iip_descriptor_t to be the same as a struct descriptor_data*, I don't
	 * need to do any explicit type casting. */
	raw_output(msg,d);
}

/* iip_error() - routes an error messag or warning to whatever logging system
 * you have in place.  Change this function to match your logging requirements.
 * */
void iip_error(gchar *msg) {
	/* my MUD game uses this simple call to lo_log(). */
	lo_log(msg);
}

/* new_iip_data() - returns a new g_malloc'ed iip_data struct, set with
 * reasonable defaults. Caller must free the returned value with
 * free_iip_data(). */
iip_data_t *new_iip_data(void) {

	iip_data_t *n;

	n = g_malloc(sizeof(iip_data_t));
	n->filename = NULL;
	n->encoded = NULL;
	n->size = 0;
	n->display = TRUE;  /* default to inline */
	n->width = 0;
	n->widthu = IIP_UNIT_NONE;
	n->height = 0;
	n->heightu = IIP_UNIT_NONE;
	n->preserveAspectRatio = TRUE;
	n->name = NULL;
	n->type = NULL;
	return(n);
}

/* free_iip_data() - g_free's any dynamically allocated values found in the
 * iip_data struct, and g_free's it too. */
void free_iip_data(iip_data_t *n) {

	if(!n) return;

	/* free any dynamically allocated strings. */
	if(n->filename) g_free(n->filename);
	if(n->encoded) g_free(n->encoded);
	if(n->type) g_free(n->type);

	g_free(n);
}

/* iip_file_get_contents() - allocates a new iip_data structure and fills it
 * with the image data stored in filename. Caller may modify the the display
 * options as needed, and must free the returned value with free_iip_data().
 * Returns NULL on error, and sends a message via iip_error(). */
iip_data_t *iip_file_get_contents(gchar *filename) {
	
	iip_data_t *n;
	gchar *contents;
	gboolean ok;
	gsize size;
	gchar *msg;

	n = new_iip_data();

	ok = g_file_get_contents(filename,&contents,&size,NULL);

	if(size > IIP_SIZE_LIMIT) {
		msg = g_strdup_printf("File '%s' is too big (%d bytes).",filename,size);
		iip_error(msg);
		g_free(msg);
		ok = FALSE;
	}

	if(!ok) {
		msg = g_strdup_printf("File '%s' didn't open?",filename);
		iip_error(msg);
		g_free(msg);
		free_iip_data(n);
		return(NULL);
	}

	n->filename = g_strdup(filename);
	n->size = size;
	n->encoded = g_base64_encode(contents,n->size);
	g_free(contents);
	return(n);
}

/* iip_output() - writes the appropriate IIP messages (via multiple calls to
 * local_binary_output()) to display the image stored in iip_data_t on the
 * terminal.  */
void iip_output(iip_data_t *image, iip_descriptor_t d) {

	gchar *msg;

	/* send the ansi OSC sequence */
	iip_raw_output("\x1b]",d);
	
	/* send IIP code */
	iip_raw_output("1337;",d);

	/* The file type is always inline- not the default of "download only" */
	iip_raw_output("File=",d);

	if(image->display == TRUE) {
		iip_raw_output("inline=1;",d);
	}

	/* send the file size. */
	msg = g_strdup_printf("size=%d;",image->size);
	iip_raw_output(msg,d);
	g_free(msg);

	/* send the width if specified */
	if(image->widthu != IIP_UNIT_NONE) {
		msg = g_strdup_printf("width=%d%s;",
			image->width,
			iip_unit_suffix[image->widthu]
		);
		iip_raw_output(msg,d);
		g_free(msg);
	}

	/* send the height if specified */
	if(image->heightu != IIP_UNIT_NONE) {
		msg = g_strdup_printf("height=%d%s;",
			image->height,
			iip_unit_suffix[image->heightu]
		);
		iip_raw_output(msg,d);
		g_free(msg);
	}

	/* only send aspect ratio setting if set to false. */
	if(image->preserveAspectRatio == FALSE) {
		iip_raw_output("preserveAspectRatio=0;",d);
	}
	
	/* name is sent b64 encoded. */
	if(image->name) {
		iip_raw_output("name=",d);
		msg = g_base64_encode(image->name,strlen(image->name));
		iip_raw_output(msg,d);
		g_free(msg);
		iip_raw_output(";",d);
	}

	/* type (mimetype) is *not* sent b64 encoded. *shrug* */
	if(image->type) {
		iip_raw_output("type=",d);
		iip_raw_output(image->type,d);
		iip_raw_output(";",d);
	}

	/* and finally the contents... */
	iip_raw_output(":",d);
	iip_raw_output(image->encoded,d);

	/* and finally the closing bell. */
	iip_raw_output("\x07",d);

	/* PHEW. */
}
